import requests
import socket
import logging
import os


def parse(filename):

    config = {}

    if os.path.exists(filename):

        with open(filename, 'r') as f:

            for line in f:
                if line[:1] == '#':
                    continue
                (key, value) = line.split('=')
                config[key.strip()] = value.strip()

    return config

def value(cfg, name, envvar=None, arg=None, default=None):
    if arg:
        return arg
    if name in cfg and cfg[name]:
        return cfg[name]
    if envvar in os.environ and os.environ[envvar]:
        return os.environ[envvar]
    return default



import boto
import os
import sh
import shutil
import logging
import requests
import hashlib
from urlparse import urlparse
from repodeploy import config


def repository(url, work_dir, *args):

    repositories = {
        's3': S3Repository,
        'http': HttpRepository,
        'https': HttpRepository,
        'git+ssh': GitRepository,
        'git+http': GitRepository,
        'git+https': GitRepository
    }

    try:
        parsed = urlparse(url)
        if parsed.scheme in repositories:
            return repositories[parsed.scheme](url, work_dir, *args)
        elif parsed.path.find('.git') > -1:
            return GitRepository(url, work_dir, *args)
    except Exception as e:
        raise Exception('Unable to create repository: %s (%s)' % (url, e))

    return None


class Repository(object):
    
    def __init__(self, url, work_dir, cfg):
        self.url = url
        self.work_dir = work_dir
        self.config = cfg
        self.log = logging.getLogger(__name__)

    def current(self, path):
        """
        Check the current configuration version from the repository to see if we need to update
        """
        raise NotImplemented()

    def fetch(self, path):
        """
        Fetch the resource at the specified path and return the local file or directory
        representing it in the working directory.
        """
        raise NotImplemented()

    def workdir(self, name, remove=False):
        d = '%s/%s' % (self.work_dir, name)
        if os.path.exists(d) and remove:
            shutil.rmtree(d)
        if not os.path.exists(d):
            os.makedirs(d)
        return d
    
class S3Repository(Repository):

    def __init__(self, url, work_dir, cfg):

        super(S3Repository, self).__init__(url, work_dir, cfg)

        parsed = urlparse(url)
        self.bucket = parsed.hostname
        self.path = parsed.path

        if 'aws-access-key' in cfg:
            self.s3 = boto.connect_s3(cfg['aws-access-key'], cfg['aws-secret-key'])
        elif os.path.exists(os.path.expanduser('~/.amazon/account-key')):
            credentials = config.parse(os.path.expanduser('~/.amazon/account-key'))
            self.s3 = boto.connect_s3(credentials['accessKey'], credentials['secretKey'])
        else:
            self.s3 = boto.connect_s3()

    def current(self):

        key = self.key(self.path)

        if key is not None:
            return key.etag.strip('"')

        return None

    def fetch(self):

        cache = self.workdir('cache', True)
        key = self.key(self.path)

        if key is not None:

            outfile = '%s/%s' % (cache, os.path.basename(key.name))

            with open(outfile, 'w') as f:
                key.get_file(f)

            unpack_dir = self.workdir('unpacked', True)
            sh.unzip(outfile, _cwd=unpack_dir)
            return (key.etag.strip('"'), unpack_dir)

            raise Exception("Unable to unpack archive format '%s'" % ext)

        return (None, cache)

    def key(self, path):

        bucket = self.s3.get_bucket(self.bucket, validate=False)

        try:
            return bucket.get_key(path)
        except:
            return None

class HttpRepository(Repository):

    def __init__(self, url, work_dir, cfg):
        super(HttpRepository, self).__init__(url, work_dir, cfg)

    def current(self):
        return self.key(requests.head(self.url))

    def key(self, response):
        if 'etag' in response.headers:
            return response.headers['etag'].strip('"')
        else:
            m = hashlib.md5()
            m.update(response.headers['content-length'])
            if 'last-modified' in response.headers:
                m.update('/')
                m.update(response.headers['last-modified'])
            return m.hexdigest()

    def fetch(self):

        cache = self.workdir('cache', True)
        r = requests.get(self.url, stream=True)

        if r.status_code == 200:

            outfile = '%s/latest.zip' % cache

            with open(outfile, 'w') as f:
                for chunk in r.iter_content(8192):
                    f.write(chunk)

            unpack_dir = self.workdir('unpacked', True)
            sh.unzip(outfile, _cwd=unpack_dir)
            return (self.key(r), unpack_dir)

            raise Exception("Unable to unpack archive format '%s'" % ext)

        return (None, cache)

class GitRepository(Repository):

    def __init__(self, url, work_dir, cfg):
        super(GitRepository, self).__init__(url, work_dir, cfg)

        if url.startswith('git+'):
            url = url[4:]

        self.remote = url
        self.prefix = ''
        self.branch = 'master'

        if not url.endswith('.git'):
            pos = url.rfind('.git') + 4
            self.remote = url[:pos]
            self.prefix = url[pos:]

        if '#' in self.prefix:
            prefix = self.prefix
            pos = prefix.rfind('#') + 1
            self.prefix = prefix[:pos]
            self.branch = prefix[pos:]

        self.local = self.workdir('git')

    def pull(self, directory):

        # Check for repository location change
        if os.path.exists('%s/%s' % (directory, '.git')):
            # Oh this will probably explode horribly at some point
            remote = str(sh.git('remote', '-v', _cwd=directory)).split('\n')[0].split('\t')[1].split(' ')[0]
            if remote != self.remote:
                # Location changed, wipe out and start again
                shutil.rmtree(directory)

        if not os.path.exists('%s/%s' % (directory, '.git')):
            # Fresh clone
            out = sh.git('clone', '-b', self.branch, '--single-branch', self.remote, directory, _cwd=os.path.dirname(directory))
            if out.exit_code != 0:
                raise Exception(str(out))
        else:
            # Ensure we revert to last good state (in case of moved files)
            out = sh.git('checkout', self.branch, _cwd=directory)
            if out.exit_code != 0:
                raise Exception(str(out))
            out = sh.git('checkout', '.', _cwd=directory)
            if out.exit_code != 0:
                raise Exception(str(out))
            # Update from remote
            out = sh.git('pull', _cwd=directory)
            if out.exit_code != 0:
                raise Exception(str(out))

        # Update any submodules
        out = sh.git('submodule', 'update', '--init', '--recursive', _cwd=directory)
        if out.exit_code != 0:
            raise Exception(str(out))

        # Parse latest revision from git
        return str(sh.git('rev-parse', 'HEAD', _cwd=directory)).strip()

    def current(self):
        return self.fetch()[0]

    def fetch(self):
        version = self.pull(self.local)
        path = '%s%s' % (self.local, self.prefix)
        if os.path.exists(path):
            return (version, path)
        return (None, None)



import sys
import os
import time
import datetime
import sched
import sh
import shutil
import logging
import subprocess
import re
import signal
from croniter import croniter
from repodeploy import repo
from dirsync import sync


class Deployer:

    def __init__(self, config):

        self.log = logging.getLogger(__name__)

        self.config = config
        self.version = None

        self.versionfile = '%s/current.version' % self.config['cache']
        self.currentdir = os.path.abspath(config['local'])
        self.workdir = '%s/work' % self.config['cache']

        # Create work dir
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

        # Create current config if nonexistent
        if not os.path.lexists(self.currentdir):
            os.makedirs(self.currentdir)
        elif os.path.exists(self.versionfile):
            with open(self.versionfile, 'r') as f:
                version = f.readline().strip()
                if version is not None:
                    self.version = version

        try:
            self.repository = repo.repository(config['remote'], self.workdir, self.config)
        except Exception as e:
            self.log.error('unable to initialize remote repository: %s (%s)' % (config['remote'], e))
            sys.exit(1)

        self.log.info('remote=%s' % re.sub(r'//[^:]*:[^@]*@', '//*****:*****@', self.repository.url))

    def run(self):

        signal.signal(signal.SIGTERM, self.handle_signal)

        scheduler = sched.scheduler(time.time, time.sleep)
        cron = croniter(self.config['schedule']) if 'schedule' in self.config else croniter('* * * * *')

        try:
            self.check_repo()
        except Exception as e:
            self.log.error("exception during update: %s" % e)

        while True:
            try:
                scheduler.enterabs(cron.get_next(float), 1, self.check_repo, ())
                scheduler.run()
            except KeyboardInterrupt as k:
                break
            except Exception as e:
                self.log.error("exception during update: %s" % e)
    
    def handle_signal(self, signo, stack):
        self.log.info('Received signal %s, shutting down' % signo)
        sys.exit(1)

    def check_repo(self):

        self.log.debug('checking for repository updates')
        version = self.repository.current()
        if version is not None:
            if version != self.version:
                self.log.info('updating to version %s' % version)
                self.update_repo()
                self.log.info('update complete')
            else:
                self.log.debug('on latest version %s' % version)
            return True
        else:
            self.log.warn('no repository found')
            return False

    def update_repo(self):

        try:

            (version, directory) = self.repository.fetch()

            if not directory:
                self.log.warn('update failed to properly unpack, skipping')
                return False

            # Pre-update scripts
            if not self.run_hooks(self.config['pre_hooks'], current=directory, previous=self.currentdir):
                self.log.warn('pre-update hooks failed, update blocked')
                return False

            save = '%s/repository.save' % self.workdir
            if os.path.exists(self.currentdir):
                self.log.debug('saving rollback to %s' % save)
                self.sync_dirs(self.currentdir, save)
            self.log.debug('syncing %s to %s' % (directory, self.currentdir))
            self.sync_dirs(directory, self.currentdir)

            self.log.debug('activated repository version: %s' % version)

            # Post-update scripts
            if not self.run_hooks(self.config['post_hooks'], current=self.currentdir, previous=save):
                if save and os.path.exists(save):
                    self.log.warn('post-update hooks failed, rollback to previous version')
                    self.sync_dirs(save, self.currentdir)
                    # Re-run post-update hooks after rollback
                    if not self.run_hooks(self.config['post_hooks'], current=self.currentdir):
                        self.log.error('post-update hooks failed after rollback, application may be unstable')
                else:
                    self.log.error('post-update hooks failed, application may be unstable')
                return False
            else:
                self.version = version
                with open(self.versionfile, 'w') as f:
                    f.write(self.version)
                return True

        except Exception as e:
            self.log.error('could not update repository: %s' % unicode(e))

        return False

    def sync_dirs(self, src, dest):
        sync(src, dest, 'sync', exclude=['^\.git$', '^\.git/.*'], logger=logging.getLogger('%s.dirsync' % __name__), create=True, purge=True)

    def run_hooks(self, hook_dir, current=None, previous=None):
        if os.path.exists(hook_dir):
            for script in os.listdir(hook_dir):
                full = '%s/%s' % (hook_dir, script)
                if os.access(full, os.X_OK):
                    self.log.info('running update hook: %s' % full)
                    # Pass in current/previous config directories
                    env = os.environ.copy()
                    env['CURRENT_CONFIG'] = current if current else ''
                    env['PREVIOUS_CONFIG'] = previous if previous else ''
                    p = subprocess.Popen(full, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
                    (out, err) = p.communicate()
                    if p.returncode > 0:
                        for line in out.split('\n'):
                            if line:
                                self.log.warn(line)
                        return False
        return True



from distutils.core import setup

setup(name='repo-deploy',
	description='Automatic code and configuration updates for continuous deployment environments',
	long_description="""
	Automates code and configuration deployment for continuous deployment environments by periodically
	polling a source repository for configuration changes (http, git, s3, etc). When changes are
	detected, the new files are retrieved and any locally configured update triggers are executed
	(in /etc/repo-deploy/*.d). This is intentionally separate from configuration management systems
	like Puppet in order to provide more flexible packaging and build process integration. The
	deployer can be configured to run as a daemon (with supervisord) or under cron (or Puppet
	agent.)
	""",
	author='Jeremy Jongsma',
	author_email='jeremy@barchart.com',
	url='https://github.com/barchart/repo-deploy',
	version='1.0.8',
	packages=['repodeploy'],
	scripts=['bin/repo-deploy'],
	data_files=[
		('/etc/repo-deploy', ['etc/repo-deploy.cfg']),
		('/etc/repo-deploy/pre-update.d', ['etc/pre-update.d/README']),
		('/etc/repo-deploy/post-update.d', ['etc/post-update.d/README'])
	],
	install_requires=['requests>=2.2.0', 'boto>=2.20.0', 'sh>=1.0', 'croniter>=0.3.5', 'dirsync>=2.1'])
